"""
Contains prompt templates for different tasks.
"""

VANILLA_QA_PROMPT = """
###
Instruction: Answer the question based on your knowledge

### QUESTION:
{question}

 """

#####
# To be used for MISTRAL model
#####

# VANILLA_QA_PROMPT = """
# ### [INST]
# Instruction: Answer the question based on your knowledge

# ### QUESTION:
# {question}

# [/INST]
#  """

# Note to add the specialized chars like [INST], [/INST] if using with a MISTRAL model
RAG_QA_PROMPT = """
###
Instruction: Answer the question based on the provided context. Only rely on
the context to build your answer and do not use your own knowledge:

{context}

### QUESTION:
{question}

"""

Q_GENERATION_PROMPT = """
###

-- INSTRUCTIONS --

You are given a CHUNK of text and a specified TOPIC_OF_INTEREST. Your job is to generate a question that has the following properties:

- The question MUST be DIRECTLY RELATED to the TOPIC_OF_INTEREST.
- The answer to the question MUST be clearly stated in the chunk.
- The question MUST not be easy to answer.
- You must not generate questions based on people names, article titles, and reference citations.

YOU MUST ONLY OUTPUT a ```JSON``` object with the key "question", and your generated question as the value for that key. If it is not possible to formulate an eligible question, return {{"question": "NONE"}}

-- EXAMPLES --

Examples of good and poor questions for an example chunk and topic of interest are provided below.

- Example topic of interest: deep learning

- Example chunk: 'Drawing boxes around the lesions, labelling of the lesions in all the slices, and training the models on the patches of detected lesions and manual labels. The time required to perform these manual operations is usually not considered when addressing the real-world application of these models and probably represents one of the major hurdles to widespread clinical adoption. A fully automatic tool running on chest CT images for the differential diagnosis of pneumonias can represent an important step forward for decreasing the variability of interpretation among clinicians and speeding up the diagnostic process. This will unburden medical staff and, in turn, provide better and faster diagnosis for patients, reducing the use of hospital resources. Better allocation of both material and human resources can be essential in a time of crisis, as the COVID-19 pandemic demonstrated with dramatic clarity [9]. To attain this goal, we developed and externally validated a fully automated deep-learning framework with a three-dimensional (3D) CNN, able to classify chest CT scans of patients with COVID-19, influenza/CAP, or no infection without manual intervention. Individual AI-based whole lung and lung abnormalities segmentation models were used to pre-process the CT images to train the 3D CNN model and are an integral part of the workflow to assure that only the patients presenting abnormalities in the lung volume are processed by the model, saving time and computational power. Material and methods'

- Example of a good question:

{{"question": "What aspects of preparing data for training deep learning models on CT scan data might be overlooked at the beginning of a deep learning project?"}}
This is a good question because the user does not need to see the chunk to understand the question, but still, the chunk provides a good example of how some requirements like annotation time could be overlooked at the beginning phase of a deep learning project.

- Example of poor questions:

{{"question": "What is the purpose of the fully automated deep-learning framework developed for chest CT scans in the context of pneumonia diagnosis?"}}
This is a poor question because a deep learning framework developed for chest CT scans in the context of pneumonia can serve many purposes, and it is not clear what framework the question is asking about.
{{"question": "What radiology modality is primarily used for diagnosing COVID 19?"}}
This is a poor question, becuase it is not relevant to the topic of interest (deep learning).
{{"question": "What is a deep learning training loop?}}
This is a poor question, because it is not based on the provided chunk!

-- TOPIC_OF_INTEREST --
{topic}

-- CHUNK --
{context}

"""

Q_EVAL_PROMPT = """
###

You are given a QUESTION and a CHUNK of text. Determine if the CHUNK directly or potentialy answers the QUESTION. Use the following instructions and examples for your evaluation:

-- INSTRUCTIONS --

- Your RESPONSE MUST be one of the two ```json``` options: {{"decision": "YES"}} or {{"decision": "NO"}}. Do not include any extra text or commentary.
- If the CHUNK is primarily composed of REFERENCE TITLES, CITATIONS, or NAMES without substantive content related to the QUESTION, you MUST respond with {{"decision": "NO"}}.
- Carefully assess whether the CHUNK explicitly mentions the answer to the QUESTION. If in doubt, err on the side of caution and consider the relevance and directness of the information.
- If the CHUNK provides detailed information related to the QUESTION but does not directly address the QUESTION's core inquiry, consider the depth and relevance of the information before deciding.
- Avoid confirmation bias; do not assume relevance based solely on related topics or keywords. The presence of keywords alone does not constitute a "YES" decision.

-- EXAMPLE 1 --

- Example QUESTION: "What are some common challenges faced during the development of computer-aided systems for detecting lung nodules on computed tomography images?"
- Example CHUNK: "Diagnostics 2021 ,11, 1405 3 of 36 Computed Tomography (CT) is often used for assessing COVID-19 severity in the lung regions and is considered an important component of the computer-aided diagnosis for lung image analysis. For a complete diagnosis of the COVID-19 severity, one must first identify the region of interest in these CT scans. There are two main challenges associated with the processing of CT scans: first and foremost, the challenge is the large volume of patients in diagnostic centers with each having 200 slices to be processed. This makes the task of processing scanned images tedious and time-consuming. The second issue with current automated or semi-automated systems is reliability, accuracy, and clinical effectiveness. One of the major causes for unreliable accuracy and low performance is the intensity-based segmentation methods which are influenced by local or global statistical methods. Furthermore, it does not take advantage of the cohort’s knowledge. Thus, there is a clear need for an automated and accurate joint left and right lung identification system in CT scans."
- explanation: The correct decision for this example is {{"decision": "YES"}} because the CHUNK explicitly discusses challenges associated with processing CT scans, which directly answers the QUESTION about challenges in developing computer-aided systems for detecting lung nodules.
- Your output MUST be this for this example: {{"decision": "YES"}}

-- QUESTION --
{question}

-- CHUNK --
{context}
"""

SYSTEM_PROMPT = "You are an expert radiologist and data scientist!"


"""
Utility functions for building a vector db. Currently Supporting ChromaDB
"""

from typing import Callable, Dict, List
import chromadb


def get_vector_db(
    db_path: str,
    db_name: str,
    embedding_fn: Callable,
    docs: List[Dict[str, str]] = [],
    db_metadata: Dict[str, str] = {'hnsw:space':'cosine'},
) -> chromadb.Collection:
    """
    A function to load or create a vector database using the provided parameters.
    If a vector database already exists, it will be returned and the docs parameter will be ignored.
    
    Args:
        db_path (str): The path to the vector database.
        db_name (str): The name of the vector database.
        embedding_fn (Callable): The function used for embedding.
        docs (List[Dict[str, str]], optional): List of documents with chunk, source, page, and index. Defaults to [].
        db_metadata (Dict[str, str], optional): Metadata for the database, default includes 'hnsw:space' as 'cosine'.
    
    Returns:
        chromadb.Collection: The vector database collection after loading or creation.
    """
    
    chroma_client = chromadb.PersistentClient(path=db_path)
    vector_db = chroma_client.get_or_create_collection(
        name=db_name,
        embedding_function=embedding_fn,
        metadata=db_metadata
    )
    if vector_db.count() > 0:
        print("Loading from disk... the given docs will be ignored.")
        return vector_db
    
    vector_db.add(
        documents=[doc["chunk"] for doc in docs],
        metadatas=[
            {
                "type": "article_chunk",
                "source": doc["source"],
                "page": doc["page"],
                "chunk_index": doc["chunk_index"]
            }
            for doc in docs
        ],
        ids=[str(i) for i in range(len(docs))]
    )

    return vector_db
    



if __name__ == "__main__":
    client = chromadb.PersistentClient(path="../data/vector_db")
    collection = client.get_or_create_collection(name="example")
    print(collection.count())



"""
Main script for RAG with conformal prediction.
"""

from typing import Dict, List, Tuple

import chromadb
from conflare.models.openai import OpenAIModelQA
from conflare.models.hf import HFModelQA
from conflare.augmented_retrieval.vector_db import get_vector_db
from conflare.augmented_retrieval.embed import TextEmbedding
from conflare.augmented_retrieval.rag import ConformalRetrievalQA
from conflare.conformal.calibration import create_calibration_records
from conflare.utils.loaders import load_dir
from conflare.utils.preprocess import chunk_docs
from conflare.utils.prompts import SYSTEM_PROMPT


def initialize_pipeline(
    document_dir: str = "./data/documents",
    db_dir: str = "./data/vector_db",
    db_name: str = "conflare",
    chunk_size: int = 1500,
    chunk_overlap: int = 200,
    model: str = "gpt-4",
    temperature: float = 0.1,
    max_tokens: int = 2000,
    frequency_penalty: float = 1.1,
    quantize: bool = True,
    embedding_model: str = "sentence-transformers/all-MiniLM-L6-v2",
    batch_size: int = 32,
    normalize_embeddings: bool = True,
    db_metadata: dict = {"hnsw:space": "cosine"},
) -> Tuple[List[Dict[str, str]], OpenAIModelQA, chromadb.Collection]:
    """
    Initializes a pipeline for processing documents. 

    Parameters:
        document_dir (str): Directory path where documents are stored. Default is "./data/documents".
        db_dir (str): Directory path where the vector database is stored. Default is "./data/vector_db".
        db_name (str): Name of the vector database. Default is "conflare".
        chunk_size (int): Size of each chunk when processing documents. Default is 1500.
        chunk_overlap (int): Amount of overlap between document chunks. Default is 200.
        model (str): Name of the model to use for question answering. Default is "gpt-4".
        temperature (float): Controls the randomness of the model. Default is 0.1.
        max_tokens (int): Maximum number of tokens to generate. Default is 2000.
        frequency_penalty (float): Adjusts the penalties for repeating tokens. Default is 1.1.
        quantize (bool): Whether to quantize the model, only used for HF models. Default is True.
        embedding_model (str): Name of the embedding model to use. Default is "sentence-transformers/all-MiniLM-L6-v2".
        batch_size (int): Number of samples per batch. Default is 32.
        normalize_embeddings (bool): Whether to normalize embeddings. Default is True.
        db_metadata (dict): Metadata for the vector database. Default is {"hnsw:space": "cosine"}.

    Returns:
        Tuple[List[Dict[str, str]], OpenAIModelQA, chromadb.Collection]: A tuple containing the processed documents, the question answering model pipeline, and the vector database.
    """
    

    docs = load_dir(document_dir)
    docs = chunk_docs(docs, chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    if "gpt" in model:
        qa_pipeline = OpenAIModelQA(
            model=model,
            system_prompt=SYSTEM_PROMPT,
            temperature=temperature,
            max_tokens=max_tokens,
            frequency_penalty=frequency_penalty,
        )
    else:
        qa_pipeline = HFModelQA(
            model=model,
            quantize=quantize,
            temperature=temperature,
            max_tokens=max_tokens,
            repetition_penalty=frequency_penalty,
        )
    embeddeing_fn = TextEmbedding(
        model_id=embedding_model,
        batch_size=batch_size,
        normalize_embeddings=normalize_embeddings,
    )
    vector_db = get_vector_db(
        db_path=db_dir,
        db_name=db_name,
        embedding_fn=embeddeing_fn,
        docs=docs,
        db_metadata=db_metadata,
    )

    return docs, qa_pipeline, vector_db


if __name__ == "__main__":
    docs, qa_pipeline, vector_db = initialize_pipeline()
    calibration_records = create_calibration_records(
        docs=docs,
        qa_pipeline=qa_pipeline,
        vector_db=vector_db,
        size=100,
        topic_of_interest="deep learning",
        max_chunk_eval=100,
        save_root="./data/calibration_set",
    )
    conformal_rag = ConformalRetrievalQA(
        qa_pipeline=qa_pipeline,
        vector_db=vector_db,
        calibration_records=calibration_records,
        error_rate=0.05,
        verbose=True,
    )

    response, retrieved_docs = conformal_rag(
        "What types of regularization methods have been used in training of the deep models?"
    )



"""
Utility functions for loading Hugging Face models and interacting with them.
"""

from typing import Tuple
from torch import cuda, bfloat16
import transformers
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    AutoConfig,
    BitsAndBytesConfig,
)


class HFModelQA:
    "Creates an object for QA using an HF model"
    def __init__(
        self,
        model: str = "mistralai/Mistral-7B-Instruct-v0.1",
        quantize: bool = True,
        temperature: float = 0.1,
        max_tokens: int = 2000,
        repetition_penalty: float = 1.1,
    ) -> None:
        """
        Initialize the model with the specified model ID or use the default one if not provided.
        
        Parameters:
            model_id (str): The ID of the model to be used.
            quantize (bool): A flag to quantize the model.
            temperature (float): The temperature for sampling the outputs.
            max_tokens (int): The maximum number of tokens to be generated.
            repetition_penalty (float): The repetition penalty to be applied.
        
        Returns:
            None
        """
        self.model, self.tokenizer = self.load_hf_model_tokenizer(model, quantize)
        self.generator = transformers.pipeline(
            model=self.model,
            tokenizer=self.tokenizer,
            task="text-generation",
            return_full_text=False,
            temperature=temperature,
            max_new_tokens=max_tokens,
            repetition_penalty=repetition_penalty,
        )
        self.generator.tokenizer.pad_token_id = self.generator.model.config.eos_token_id

    def load_hf_model_tokenizer(
        self, model_id: str, quantize: bool
    ) -> Tuple[AutoModelForCausalLM, AutoTokenizer]:
        """
        Load a model and tokenizer for language modeling.

        Args:
            model_id (str): The ID of the model to be loaded.
            quantize (bool): Whether to quantize the model.

        Returns:
            Tuple[AutoModelForCausalLM, AutoTokenizer]: The loaded model and its corresponding tokenizer.
        """
        device = f"cuda: {cuda.current_device()}" if cuda.is_available() else "cpu"
        bnb_config = None
        if quantize:
            bnb_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_use_double_quant=True,
                bnb_4bit_compute_dtype=bfloat16,
            )
        model_config = AutoConfig.from_pretrained(model_id)
        model = AutoModelForCausalLM.from_pretrained(
            model_id,
            config=model_config,
            device_map="auto",
            trust_remote_code=True,
            quantization_config=bnb_config,
        )
        model.eval()
        print(f"Model loaded on {device} successfully.")

        tokenizer = AutoTokenizer.from_pretrained(model_id)

        return model, tokenizer

    def __call__(self, prompt: str) -> str:
        """
        Call the generator with the given prompt and return the generated text.

        Parameters:
            prompt (str): The prompt to be passed to the generator.

        Returns:
            str: The generated text stripped of leading and trailing whitespace.
        """
        response = self.generator(prompt, pad_token_id=self.tokenizer.eos_token_id)

        return response[0]["generated_text"].strip()


# def qa_with_hf_models(
#     prompt: str,
#     model: AutoModelForCausalLM,
#     tokenizer: AutoTokenizer,
#     temperature: float = 0.1,
#     max_tokens: int = 2000,
#     repetition_penalty: float = 1.1,
# ) -> str:
#     """
#     Generates a response using a Hugging Face model for question answering.

#     Args:
#         prompt (str): The prompt for the question answering.
#         model (AutoModelForCausalLM): The Hugging Face model for question answering.
#         tokenizer (AutoTokenizer): The Hugging Face tokenizer for question answering.
#         temperature (float, optional): The temperature for sampling. Defaults to 0.1.
#         max_tokens (int, optional): The maximum number of tokens to generate. Defaults to 2000.
#         repetition_penalty (float, optional): The repetition penalty to apply. Defaults to 1.1.

#     Returns:
#         str: The generated response.
#     """

#     generator = transformers.pipeline(
#         model=model,
#         tokenizer=tokenizer,
#         task="text-generation",
#         return_full_text=False,
#         temperature=temperature,
#         max_new_tokens=max_tokens,
#         repetition_penalty=repetition_penalty,
#     )
#     generator.tokenizer.pad_token_id = generator.model.config.eos_token_id
#     response = generator(prompt, pad_token_id=tokenizer.eos_token_id)

#     return response[0]["generated_text"].strip()



"""
Functions for building a calibration set to be used for conformal prediction.
"""

import os
import pickle
from typing import Callable, Dict, List
import random
import chromadb
from tqdm import tqdm

from conflare.utils.prompts import Q_GENERATION_PROMPT, Q_EVAL_PROMPT


class QuestionGeneration:
    "A class for generating questions based on chunks and saving / loading them to / from a pickle file."

    def __init__(
        self,
        docs: List[Dict[str, str]] | None = None,
        qa_pipeline: Callable | None = None,
        num_questions: int | None = 100,
        topic_of_interest: str | None = None,
        q_generation_prompt: str = Q_GENERATION_PROMPT,
        path_to_pickle: str | None = None,
    ) -> None:
        """
        Args:
            docs (List[Dict[str, str]] | None, optional): A list of dictionaries representing documents. Defaults to None.
            qa_pipeline (Callable | None, optional): A Callable object representing the question-answering pipeline. Defaults to None.
            num_questions (int | None, optional): The number of questions to generate. Defaults to 100.
            topic_of_interest (str | None, optional): The topic of interest for the questions. Defaults to None.
            q_generation_prompt (str, optional): The prompt for generating questions. Defaults to Q_GENERATION_PROMPT.
            path_to_pickle (str | None, optional): The path to a pickle file. Defaults to None.
        """

        self.docs = docs
        self.qa_pipeline = qa_pipeline
        self.num_questions = num_questions
        self.topic_of_interest = topic_of_interest
        self.q_generation_prompt = q_generation_prompt
        self.path_to_pickle = path_to_pickle
        self.generated_questions = []
        self.error_logs = []

    def generate_questions(self, save_to_disk: bool = False) -> List[List]:
        """
        Generate questions based on the content of the documents (chunks) using the qa_pipeline.

        Parameters:
            save_to_disk (bool): Whether to save the generated questions to disk. Default is True.

        Returns:
            None
        """
        random.shuffle(self.docs)
        generation_pbar = tqdm(total=self.num_questions, desc="Generation Progress")
        for i, doc in enumerate(self.docs):
            chunk = doc["chunk"]
            prompt = self.q_generation_prompt.format(topic=self.topic_of_interest, context=chunk)
            response = self.qa_pipeline(prompt)

            try:
                g_question = eval(response)["question"]
            except:
                self.error_logs.append([i, doc, response])
                continue

            if g_question.lower() != "none":
                self.generated_questions.append([i, doc, g_question])
                generation_pbar.update(1)

            if len(self.generated_questions) == self.num_questions:
                break

        if save_to_disk:
            self.save()
        
        return self.generated_questions

    def save(self):
        """
        Saves the generated questions and related information to a pickle file.

        This function creates a dictionary called `generated_questions_records` that contains the following keys:
        - 'model': The model used for generating the questions.
        - 'topic_of_interest': The topic of interest for generating the questions.
        - 'generated_questions': The list of generated questions.

        The `generated_questions_records` dictionary is then serialized and saved to a pickle file specified by `self.path_to_pickle`.

        Returns:
            None
        """
        generated_questions_records = {
            "Model": self.qa_pipeline.model,
            "topic_of_interest": self.topic_of_interest,
            "generated_qs": self.generated_questions,
        }

        with open(self.path_to_pickle, "wb") as f:
            pickle.dump(generated_questions_records, f)
        print("Saved to disk!")

    @classmethod
    def from_pickle(cls, path_to_pickle: str = None):
        """
        Load data from a pickle file and return a class instance with the relevant values.

        Parameters:
            path_to_pickle (str): The path to the pickle file to load.

        Raises:
            FileNotFoundError: If the specified file is not found.
        """
        if os.path.exists(path_to_pickle):
            with open(path_to_pickle, "rb") as f:
                generated_questions_records = pickle.load(f)

            instance = cls(
                topic_of_interest=generated_questions_records["topic_of_interest"],
                num_questions=len(generated_questions_records["generated_qs"]),
            )
            instance.generated_questions = generated_questions_records["generated_qs"]
            return instance

        else:
            raise FileNotFoundError(f"File {path_to_pickle} not found.")
    
    def get_generated_questions(self) -> List[List]:
        return self.generated_questions

    def __getitem__(self, idx: int):
        if len(self.generated_questions) != 0:
            return self.generated_questions[idx][-1]
        else:
            raise IndexError(
                "There are no generated questions. Please generate some questions first using the generate_questions() method or load the data from the previously saved pickle file using the load() method."
            )

    def __len__(self):
        return len(self.generated_questions)


class QuestionEvaluation:
    def __init__(
        self,
        questions: List[list] | None = None,
        vector_db: chromadb.Collection | None = None,
        qa_pipeline: Callable | None = None,
        eval_prompt: str = Q_EVAL_PROMPT,
        max_chunk_eval: int = 100,
        path_to_pickle: str | None = None,
    ) -> None:
        """
        Args:
            questions (List[list]): A list of questions to be used for initialization.
            vector_db (chromadb.Collection): The vector database for the QA pipeline.
            qa_pipeline (Callable): The QA pipeline for evaluating the questions against chunks.
            eval_prompt (str, optional): The evaluation prompt. Defaults to Q_EVAL_PROMPT.
            max_chunk_eval (int, optional): The maximum number of retrieved chunks to evaluate. Defaults to 100.
            path_to_pickle (str, optional): The path to the pickle file to load or save.

        Returns:
            None
        """

        self.questions = questions
        self.vector_db = vector_db
        self.qa_pipeline = qa_pipeline
        self.eval_prompt = eval_prompt
        self.max_chunk_eval = max_chunk_eval
        self.path_to_pickle = path_to_pickle
        self.calibration_records = []

    def evaluate_questions(self, save_to_disk: bool = False) -> List[Dict[str, str]]:
        for i, question_list in tqdm(enumerate(self.questions)):
            _, doc, question = question_list
            source_chunk = doc["chunk"]
            retrieved_chunks = self.vector_db.query(
                query_texts=question,
                n_results=self.vector_db.count(),
            )
            sorted_chunks = retrieved_chunks["documents"][0]
            sorted_scores = retrieved_chunks["distances"][0]
            source_chunk_index = sorted_chunks.index(source_chunk)

            relevant_chunk_found = False
            max_chunk_eval = min(self.max_chunk_eval, len(sorted_chunks))

            for j, (chunk_text, chunk_score) in enumerate(
                zip(sorted_chunks[:max_chunk_eval], sorted_scores[:max_chunk_eval])
            ):

                if chunk_text.lower() == source_chunk.lower():
                    relevant_chunk_found = True
                    break

                else:
                    response = self.qa_pipeline(
                        self.eval_prompt.format(question=question, context=chunk_text)
                    )
                    try:
                        decision = eval(response)["decision"]
                    except:
                        continue

                    if decision.lower() == "yes":
                        relevant_chunk_found = True
                        break

            if relevant_chunk_found:
                record = {
                    "generated_q": question_list,
                    "source_chunk_text": source_chunk,
                    "source_chunk_rank": source_chunk_index,
                    "relevant_chunk_text": chunk_text,
                    "relevant_chunk_rank": j,
                    "cosine_distance": chunk_score,
                }
                self.calibration_records.append(record)

        if save_to_disk:
            self.save()
        
        return self.calibration_records

    def save(self):
        with open(self.path_to_pickle, "wb") as f:
            pickle.dump(self.calibration_records, f)
        print("Saved to disk!")

    @classmethod
    def from_pickle(cls, path_to_pickle: str = None):
        if os.path.exists(path_to_pickle):
            with open(path_to_pickle, "rb") as f:
                calibration_records = pickle.load(f)

            instance = cls(questions=[record["generated_q"] for record in calibration_records])
            instance.calibration_records = calibration_records
            return instance

        else:
            raise FileNotFoundError(f"File {path_to_pickle} not found.")

    def get_calibration_records(self) -> List[Dict[str, str]]:
        return self.calibration_records

    def __getitem__(self, idx: int):
        if len(self.calibration_records) != 0:
            return self.calibration_records[idx]
        else:
            raise IndexError(
                "There are no calibration records. Please evaluate some questions first using the evaluate() method or load the data from the previously saved pickle file using the load() method."
            )

    def __len__(self):
        return len(self.calibration_records)


def create_calibration_records(
    docs: List[Dict[str, str]],
    qa_pipeline: Callable,
    vector_db: chromadb.Collection,
    size: int,
    topic_of_interest: str,
    max_chunk_eval: int,
    generation_prompt: str = Q_GENERATION_PROMPT,
    evaluation_prompt: str = Q_EVAL_PROMPT,
    save_root: str = "./data/calibration_set",
    save_to_disk: bool = False,
) -> List[Dict[str, str]]:
    """
    Create calibration records for a given set of documents.

    Args:
        docs (List[Dict[str, str]]): A list of documents to create calibration records for.
        size (int): The size of the final calibration set (calibration records).
        topic_of_interest (str): The topic of interest for down stream RAG QA.
        qa_pipeline (Callable): The question-answering pipeline to use for generating questions.
        vector_db (chromadb.Collection): The vector db to use for evaluating questions.
        save_root (str): The root directory where the generated questions and calibration records will be saved.
        save_to_disk (bool, optional): Whether to save the generated questions and calibration records to disk. Defaults to False.

    Returns:
        calibration_records: The generated calibration records.
    """
    q_generation = QuestionGeneration(
        docs=docs,
        qa_pipeline=qa_pipeline,
        num_questions=size,
        topic_of_interest=topic_of_interest,
        q_generation_prompt=generation_prompt,
        path_to_pickle=os.path.join(save_root, "Generated_Questions.pkl"),
    )
    generated_questions = q_generation.generate_questions(save_to_disk=save_to_disk)
    q_evaluation = QuestionEvaluation(
        questions=generated_questions,
        qa_pipeline=qa_pipeline,
        vector_db=vector_db,
        eval_prompt=evaluation_prompt,
        max_chunk_eval=max_chunk_eval,
        path_to_pickle=os.path.join(save_root, "Calibration_Records.pkl"),
    )
    calibration_records = q_evaluation.evaluate_questions(save_to_disk=save_to_disk)

    return calibration_records



"""
Embedding utilities for augmented retrieval.
"""

from typing import List
import torch
from sentence_transformers import SentenceTransformer


class TextEmbedding:
    """
    A wrapper for sentece transformers to get embeddings for input text
    """

    def __init__(
        self,
        model_id: str = "sentence-transformers/all-MiniLM-L6-v2",
        batch_size: int = 32,
        normalize_embeddings: bool = True,
    ) -> None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = SentenceTransformer(model_id, device=device)
        self.batch_size = batch_size
        self.normalize_embeddings = normalize_embeddings

    def __call__(self, input: List[str]) -> List[List[float]]:
        embeddings = self.model.encode(
            input, batch_size=self.batch_size, normalize_embeddings=self.normalize_embeddings
        ).tolist()

        return embeddings




"""
Utility functions for interacting with OpenAI API.
"""

import os
import openai


def setup_openai_api(api_key: str | None = None) -> None:
    """
    A function to set up the OpenAI API by providing an API key,
    either through parameter or environment variable.

    Parameters:
    api_key (str | None): The API key to authenticate with OpenAI API. Defaults to None.

    Returns:
    None
    """

    if api_key is not None:
        os.environ["OPENAI_API_KEY"] = api_key
        openai.api_key = api_key
    elif os.getenv("OPENAI_API_KEY") is not None:
        openai.api_key = os.getenv("OPENAI_API_KEY")
    else:
        raise ValueError(
            "No OpenAI API key provided. Please set it as an env variable in the os.environ"
        )


class OpenAIModelQA:
    "Creates an object for QA using an OpenAI model"

    def __init__(
        self,
        model: str = "gpt-4",
        system_prompt: str = "",
        temperature: float = 0.1,
        max_tokens: int = 2000,
        frequency_penalty: float = 1.1,
        api_key: str | None = None,
    ) -> None:
        """
        Initializes the class with the specified model, system prompt, temperature, max tokens, frequency penalty, and API key.

        Args:
            model (str): The GPT model to use for the generation.
            system_prompt (str): The prompt to provide to the GPT model.
            temperature (float): The temperature parameter for generation.
            max_tokens (int): The maximum number of tokens to generate.
            frequency_penalty (float): The frequency penalty parameter for generation.
            api_key (str | None): The API key for accessing the OpenAI API, or None if not provided.
        """

        setup_openai_api(api_key)
        self.model = model
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.frequency_penalty = frequency_penalty
        self.system_prompt = system_prompt
        self.client = openai.OpenAI(api_key=os.getenv("OPENAI_API_KEY"))

    def __call__(self, prompt: str) -> str:
        """
        A method that generates a response based on the given prompt.

        Parameters:
            prompt (str): The prompt for which a response needs to be generated.

        Returns:
            str: The response generated based on the prompt.
        """

        messages = [
            {"role": "system", "content": self.system_prompt},
            {"role": "user", "content": prompt},
        ]
        response = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            frequency_penalty=self.frequency_penalty,
        )

        return response.choices[0].message.content.strip()


# def qa_with_openai_models(
#     prompt: str,
#     model: str = "gpt-4",
#     system_prompt: str = "",
#     temperature: float = 0.1,
#     max_tokens: int = 2000,
#     frequency_penalty: float = 1.1,
# ) -> str:
#     """
#     Basic function for QA wih OpenAI models.

#     Args:
#         prompt (str): The prompt for the question-answering task.
#         model (str, optional): The name of the OpenAI model to use. Defaults to "gpt-4".
#         system_prompt (str, optional): The system prompt to include in the conversation. Defaults to "".
#         temperature (float, optional): The temperature parameter for sampling responses. Defaults to 0.1.
#         max_tokens (int, optional): The maximum number of tokens in the response. Defaults to 2000.
#         frequency_penalty (float, optional): The frequency penalty parameter for sampling responses. Defaults to 1.1.

#     Returns:
#         str: The generated response.
#     """

#     messages = [{"role": "system", "content": system_prompt}, {"role": "user", "content": prompt}]
#     client = openai.OpenAI(api_key=os.getenv("OPENAI_API_KEY"))

#     response = client.chat.completions.create(
#         model=model,
#         messages=messages,
#         temperature=temperature,
#         max_tokens=max_tokens,
#         frequency_penalty=frequency_penalty,
#     )

#     return response.choices[0].message.content.strip()



"""
Conflare: Conformal Retreival for Augmented Generation with LLMs
"""
import conflare.augmented_retrieval
import conflare.conformal
import conflare.utils
import conflare.models

from .main import initialize_pipeline

__version__ = "0.1.0"



"""
Retrieval Augmented Generation (RAG) utilities.
"""

from typing import Callable, Dict, List
import chromadb
import numpy as np

from conflare.utils.prompts import RAG_QA_PROMPT


class SimpleRetrievalQA:
    "Creates an object to do RAG based QA using an HF or OpenAI model"

    def __init__(self, qa_pipeline: Callable, vector_db: chromadb.Collection, topk: int) -> None:
        """
        Args:
            qa_pipeline (Callable): The QA pipeline to be used; can be HF or OpenAI Callable pipeline.
            vector_db (chromadb.Collection): The collection of vectors for the database.
            topk (int): The number of documents to retrieve.
        """
        self.vector_db = vector_db
        self.qa_pipeline = qa_pipeline
        self.topk = topk

    def retrieve_docs(self, query: str, topk: int) -> List[str]:
        retrieved_docs = self.vector_db.query(query_texts=query, n_results=topk)["documents"][0]
        return retrieved_docs

    def __call__(self, question: str, retuen_retrieved_chunks: bool = True) -> str:
        """
        A function that takes a question, retrieves documents, generates context, creates a prompt,
        and gets a response from a QA pipeline. Returns the response and retrieved documents based on the flag.

        Parameters:
            question (str): The question to be answered.
            topk (int): Number of documents to retrieve (default is 3).
            retuen_retrieved_chunks (bool): Flag to indicate whether to return retrieved documents along with the response (default is True).

        Returns:
            str: The response generated by the QA pipeline.
            list: Retrieved documents based on the flag.
        """
        retrieved_docs = self.retrieve_docs(query=question, topk=self.topk)
        context = "\n\n".join(retrieved_docs)
        prompt = RAG_QA_PROMPT.format(context=context, question=question)
        response = self.qa_pipeline(prompt)
        if retuen_retrieved_chunks:
            return response, retrieved_docs
        else:
            return response


class ConformalRetrievalQA(SimpleRetrievalQA):
    "Creates an object to do RAG based QA using an HF or OpenAI model with conformal filtering"
    def __init__(
        self,
        qa_pipeline: Callable,
        vector_db: chromadb.Collection,
        calibration_records: List[Dict[str, str]],
        error_rate: float = 0.05,
        verbose: bool = False,
    ) -> None:
        self.calibration_records = calibration_records
        self.error_rate = error_rate
        self.verbose = verbose
        super().__init__(qa_pipeline, vector_db, topk=vector_db.count())

    def retrieve_docs(self, query: str, topk: int) -> List[str]:
        retrieved_docs = self.vector_db.query(query_texts=query, n_results=topk)
        return self.conformal_filter(retrieved_docs)

    def conformal_filter(self, retrieved_docs: Dict[str, list]):
        """
        Filters the retrieved documents based on a calibration records.

        Args:
            retrieved_docs (Dict[str, list]): A dictionary containing the retrieved documents.
                The dictionary should have two keys: "documents" and "distances".
                The value corresponding to the "documents" key is a list of document texts.
                The value corresponding to the "distances" key is a list of cosine distances.

        Returns:
            List[str]: A list of relevant chunks from the retrieved documents.
                The list contains the document texts that have a cosine distance less than the threshold.
        """
        cosine_distances = [record["cosine_distance"] for record in self.calibration_records]
        threshold = np.percentile(cosine_distances, 100 - self.error_rate * 100)
        doc_texts = retrieved_docs["documents"][0]
        doc_distances = retrieved_docs["distances"][0]
        relevant_docs = [
            doc for doc, distance in zip(doc_texts, doc_distances) if distance < threshold
        ]
        if self.verbose:
            print(f"Input Error Rate: {self.error_rate * 100:.2f}%")
            print(f"Selected cosine distance thereshold: {threshold:.3f}")
            print(f"Number of retrieved documents: {len(relevant_docs)}")

        return relevant_docs



"""
Pre-processing and text cleaning utilities.
"""

from typing import Dict, List


def chunk_text(text: str, chunk_size: int, chunk_overlap: int) -> List[str]:
    """
    A function that chunks a text into smaller pieces with a specified size and overlap.

    Parameters:
    - text (str): The input text to be chunked.
    - chunk_size (int): The size of each chunk.
    - chunk_overlap (int): The amount of overlap between consecutive chunks.

    Returns:
    - List[str]: A list of chunked text segments.
    """

    chunked_docs = []
    i = 0
    while i < len(text):
        end_index = min(i + chunk_size, len(text))
        chunk = text[i: end_index]
        chunked_docs.append("..." + chunk + "...")
        i += chunk_size - chunk_overlap
        if i + chunk_size - chunk_overlap > len(text):
            break

    return chunked_docs


def chunk_docs(
    docs: List[Dict[str, str]], chunk_size: int = 1500, chunk_overlap: int = 200
) -> List[Dict[str, str]]:
    """
    Generates chunks of text from a list of documents.

    Args:
        docs (List[Dict[str, str]]): A list of documents, where each document is represented as a dictionary with "content" and other optional fields.
        chunk_size (int, optional): The maximum size of each chunk. Defaults to 1500.
        chunk_overlap (int, optional): The number of characters by which each chunk should overlap. Defaults to 200.

    Returns:
        List[Dict[str, str]]: A list of dictionaries, where each dictionary represents a chunk of text. 
        Each dictionary contains the chunk's text in "chunk", the index of the chunk in "chunk_index" within the document, 
        and any other fields from the original document that are not "content".
    """
    
    chunked_docs = []
    for doc in docs:
        chunks = chunk_text(doc["content"], chunk_size, chunk_overlap)
        for i, chunk in enumerate(chunks):
            chunked_docs.append(
                {"chunk": chunk, "chunk_index": i, **{k: v for k, v in doc.items() if k != "content"}}
            )

    return chunked_docs

def remove_illegal_chars(text: str) -> str:
    """
    Removes illegal characters from the given text.

    Args:
        text (str): The text from which illegal characters need to be removed.

    Returns:
        str: The text with illegal characters removed.
    """
    illegal_chars = [
        "\x00",
        "\x01",
        "\x02",
        "\x03",
        "\x04",
        "\x05",
        "\x06",
        "\x07",
        "\x08",
        "\x0b",
        "\x0c",
        "\x0e",
        "\x0f",
        "\x10",
        "\x11",
        "\x12",
        "\x13",
        "\x14",
        "\x15",
        "\x16",
        "\x17",
        "\x18",
        "\x19",
        "\x1a",
        "\x1b",
        "\x1c",
        "\x1d",
        "\x1e",
        "\x1f",
    ]
    for char in illegal_chars:
        text = text.replace(char, "")

    return text



"""
Document loaders for different formats.
"""

from typing import Dict, List
from glob import glob
from pypdf import PdfReader
from tqdm import tqdm

from conflare.utils.preprocess import remove_illegal_chars


def pdf_loader(pdf_path: str) -> List[Dict[str, str]]:
    """
    Load a PDF file and extract its contents into a list of dictionaries.

    Parameters:
        pdf_path (str): The path to the PDF file.

    Returns:
        List[Dict[str, str]]: list of dicts, where each dictionary contains the source PDF path,
        the page number, and the extracted text content of each page in the PDF file.
    """
    pdf_docs = []
    reader = PdfReader(pdf_path)
    for i, page in enumerate(reader.pages):
        content = remove_illegal_chars(page.extract_text())
        pdf_docs.append({"source": pdf_path, "page": i, "content": content})

    return pdf_docs


def load_dir(path: str, extension: str = "pdf") -> List[Dict[str, str]]:
    """
    Loads all files with a specified extension from a given directory and returns
    a list of dictionaries representing the documents.

    Parameters:
        path (str): The path to the directory containing the files.
        extension (str, optional): The file extension to filter the files by. Defaults to "pdf".

    Returns:
        List[Dict[str, str]]: A list of dictionaries representing the documents.
        Each dictionary contains the document path and its content.

    Note:
        The function currently supports loading PDF files. Other formats are not implemented.
    """
    doc_paths = glob(f"{path}/*.{extension}")
    print(f"Found {len(doc_paths)} {extension} files in {path}.")
    docs = []
    if extension == "pdf":
        loader = pdf_loader
    else:
        # TODO: Implement other formats
        raise NotImplementedError(f"Extension {extension} not supported")

    for doc_path in tqdm(doc_paths):
        docs.extend(loader(doc_path))

    return docs



from setuptools import setup

with open("./README.md", "r") as f:
    long_description = f.read()

setup(
    name='conflare',
    version='0.1.3',    
    description='conformal retreival augmented generation with LLMs',
    url='https://github.com/Mayo-Radiology-Informatics-Lab/conflare',
    author='Moein Shariatnia',
    author_email='moein.shariatnia@gmail.com',
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir = {"conflare": "conflare",
                   "conflare.conformal": "conflare/conformal",
                   "conflare.models": "conflare/models",
                   "conflare.utils": "conflare/utils",
                   "conflare.augmented_retrieval": "conflare/augmented_retrieval"},
    packages = ['conflare', 'conflare.conformal', 'conflare.models', 'conflare.utils', 'conflare.augmented_retrieval'],
    install_requires=[  
                        'pypdf==4.1.0',
                        'torch>=2.0.0',
                        'transformers==4.37.2',
                        'bitsandbytes==0.42.0',
                        'accelerate==0.26.1',
                        'openai==1.12.0',
                        'sentence-transformers==2.3.1',
                        'chromadb==0.4.22'
                    ],

    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)